import Foundation


extension Array {
    func permutations() -> [[Element]] {
        if count <= 1 {
            return [self]
        }
        
        var permutations = [[Element]]()
        
        for (index, element) in self.enumerated() {
            var remaining = self
            remaining.remove(at: index)
            let subPermutations = remaining.permutations()
            
            for var subPermutation in subPermutations {
                subPermutation.insert(element, at: 0)
                permutations.append(subPermutation)
            }
        }
        
        return permutations
    }
}


func zaehle_buchstabenhauefigkeiten(_ text: String) -> [Character: Int] {
    var frequencies = [Character: Int]()
    let letters = text.filter { $0.isLetter }.map { Character($0.uppercased()) }
    for letter in letters {
      //TODO: Häufigkeiten zählen
    }
    return frequencies
}

let buchstabenhaeufigkeitenAbsteigend = "ENISRTHADULCMOGWBFKPZVJYXQ"

class CaesarSimple {
    func verschiebeBuchstabe(_ ch: Character, _ abstand: Int) -> Character {
    var code = Int(ch.asciiValue!)
    code += abstand
    if code > 90 {
    code -= 26
    }
    if code < 65 {
    code += 26
    }
    let result = Character(UnicodeScalar(code)!)
    return result
    }

    func verschiebeSatz(_ text: String, _ abstand: Int) -> String {
        var verschluesselt = ""
        for ch in text {
            if ch.isLetter {
                let verschoben = verschiebeBuchstabe(ch, abstand)
                verschluesselt.append(verschoben)
            } else {
                verschluesselt.append(ch)
            }
        }
        return verschluesselt
    }

    func verschluesseln(_ ciphertext: String, _ schluessel: Int) -> String {
        return verschiebeSatz(ciphertext.uppercased(), schluessel)
    }

    func entschluesseln(_ ciphertext: String, _ schluessel: Int) -> String {
        let negierterSchluessel = -schluessel
        return verschiebeSatz(ciphertext.uppercased(), negierterSchluessel)
    }

    func bruteforceCaesar(_ ciphertext: String) -> [String] {
        var klartexte = [String]()
        for schluessel in 0..<26 {
           // TODO: Versuchen ciphertext mit schluessel zu knacken
        }
        return klartexte
    }

    func bruteforceCaesarErgaenztUmAbbruchbedingung(_ ciphertext: String) -> String? {
        var klartexte = [String]()
        for schluessel in 0..<26 {
            // TODO: Versuchen ciphertext mit schluessel zu knacken
            // TODO: Abbruchbedingung prüfen und Ergebnis zurückgeben

        }
        return nil
    }

    func verschluesselnBeliebigePermutation(_ satz: String, _ schluessel: [Character: Character]) -> String {
        var ciphertext = ""
        for c in satz.uppercased() {
            if let verschoben = schluessel[c] {
                ciphertext.append(verschoben)
            } else {
                ciphertext.append(c)
            }
        }
        return ciphertext
    }

    func entschluesselnBeliebigePermutation(_ ciphertext: String, _ schluessel: [Character: Character]) -> String {
        let invSchluessel = Dictionary(uniqueKeysWithValues: schluessel.map { ($1, $0) })
        return verschluesselnBeliebigePermutation(ciphertext, invSchluessel)
    }

    func bruteforceCaesarBeliebigeZuordnung(_ ciphertext: String) -> String? {
        let klartextAlphabet = Array(buchstabenhaeufigkeitenAbsteigend)
        let permutations = klartextAlphabet.permutations()
        for perm in permutations {
            let geheimtextAlphabet = perm
            let schluessel = Dictionary(uniqueKeysWithValues: zip(klartextAlphabet, geheimtextAlphabet))
            // TODO: ciphertext mit schluessel entschluesseln
            // TODO: Abbruchbedingung prüfen und gegebenenfalls Klartext zurückgeben
            
        }
        return nil
    }

    func findeSchluesselDurchVergleichDerBuchstabenhauefigkeiten(_ ciphertext: String) -> [Character: Character] {
        // Hier gibt es nichts zu tun, sie müssen lediglich die Methode zaehle_buchstabenhauefigkeiten implementieren
        let haeufigkeiten = zaehle_buchstabenhauefigkeiten(ciphertext)
        let haeufigkeitenAbsteigendSortiertNurBuchstaben = buchstabenhaeufigkeitenAbsteigend.filter { haeufigkeiten.keys.contains($0) }
        
        print(haeufigkeiten)
        let sortedHaeufigkeiten = haeufigkeiten.sorted { $0.value > $1.value }
        
        var schluessel = [Character: Character]()
        
        for (index, (key, _)) in sortedHaeufigkeiten.enumerated() {
            if index < haeufigkeitenAbsteigendSortiertNurBuchstaben.count {
                let stringIndex = haeufigkeitenAbsteigendSortiertNurBuchstaben.index(haeufigkeitenAbsteigendSortiertNurBuchstaben.startIndex, offsetBy: index)
                let character = haeufigkeitenAbsteigendSortiertNurBuchstaben[stringIndex]
                schluessel[character] = key
            }
        }
        
        return schluessel
    }
}

let caesar = CaesarSimple()
let verschluesselterText = caesar.verschluesseln("Hi, dies ist ein Text", 3)
print("-----------Brute Force    --------------")

print(caesar.bruteforceCaesar(verschluesselterText))
print(caesar.bruteforceCaesarErgaenztUmAbbruchbedingung(verschluesselterText))
print("-----------Buchstabenhäufigkeiten    --------------")

let fileURL = URL(fileURLWithPath: "./faust.txt")

do {
    let langerTextUnverschluesselt = try String(contentsOf: fileURL, encoding: .utf8)
    let langerText = caesar.verschluesseln(langerTextUnverschluesselt.uppercased(), 7)
    print("Verschlüsselter Text: ", String(langerText.prefix(100)))


    let schluessel = caesar.findeSchluesselDurchVergleichDerBuchstabenhauefigkeiten(langerText)
    print(schluessel)

    print("Klartext: ", caesar.entschluesselnBeliebigePermutation(langerText, schluessel).prefix(100))

} catch {
    print("Failed to read the file:", error)
}

